@@ -3,7 +3,7 @@ from .support import TDLiveDataError, TDHistoricDataError
 from .TD_live import LiveClient
 from .TD_hist import HistoricalREST
 from threading import Thread
-from typing import Union
+from typing import Union, Callable, Any
 from collections import defaultdict
 
 from datetime import datetime
@@ -23,11 +23,11 @@ class TD:
                  broker_token=None,
                  url='push.truedata.in',  # This arg is used for LIVE URL only
                  live_port=8082,
-                 historical_port=8092,
-                 logger_propagate=False,
+                 historical_api=True,
                  # tz='Asia/Kolkata',
-                 logger_level=20,
-                 logger=None,
+                 log_level=logging.WARNING,
+                 log_handler=None,
+                 log_format=None,
                  hist_url='https://history.truedata.in'):
         self.live_websocket = None
         self.historical_datasource = None
@@ -36,21 +36,26 @@ class TD:
         self.live_url = url
         self.hist_url = hist_url
         self.live_port = live_port
-        self.historical_port = historical_port
-        if logger is None:
-            self.logger = logging.getLogger(__name__)
-            self.logger.propagate = logger_propagate
-            self.logger.setLevel(logger_level)
+        self.connect_historical = historical_api
+        if log_format is None:
+            # log_format = "%(levelname)s : %(message)s"
+            log_format = "(%(asctime)s) %(levelname)s :: %(message)s (PID:%(process)d Thread:%(thread)d)"
+        if log_handler is None:
+            log_formatter = logging.Formatter(log_format)
+            # log_formatter = logging.Formatter("%(message)s")
+            self.log_handler = logging.StreamHandler()
+            self.log_handler.setLevel(log_level)
+            self.log_handler.setFormatter(log_formatter)
         else:
-            self.logger = logger
+            self.log_handler = log_handler
+        self.logger = logging.getLogger(__name__)
+        self.logger.addHandler(self.log_handler)
+        self.logger.setLevel(self.log_handler.level)
+        self.logger.debug("Logger ready...")
         if live_port is None:
             self.connect_live = False
         else:
             self.connect_live = True
-        if historical_port is None:
-            self.connect_historical = False
-        else:
-            self.connect_historical = True
         self.broker_token = broker_token
         self.hist_data = {}
         self.live_data = {}
@@ -59,7 +64,6 @@ class TD:
         # self.streaming_symbols = {}
         self.touchline_data = {}
         self.default_market_data_id = 2000
-        self.default_historical_data_id = 1000
         self.connect()
 
     def connect(self):
@@ -71,7 +75,7 @@ class TD:
             t = Thread(target=self.connect_thread, args=(), daemon=True)
             t.start()
         if self.connect_historical:
-            self.historical_datasource = HistoricalREST(self.login_id, self.password, self.hist_url, self.historical_port, self.broker_token, self.logger)
+            self.historical_datasource = HistoricalREST(self.login_id, self.password, self.hist_url, self.broker_token, self.logger)
         while self.connect_live and self.live_websocket.subscription_type == '':
             time.sleep(1)
 
@@ -84,10 +88,10 @@ class TD:
             self.live_websocket.disconnect_flag = True
             self.live_websocket.close()
             # self.logger.info(f"{Fore.GREEN}Disconnected LIVE TrueData...{Style.RESET_ALL}")
-            print(f"{Style.NORMAL}{Fore.BLUE}Disconnected from Real Time Data WebSocket Connection !{Style.RESET_ALL}")
+            self.logger.warning(f"{Style.NORMAL}{Fore.BLUE}Disconnected from Real Time Data WebSocket Connection !{Style.RESET_ALL}")
         if self.connect_historical:
             # self.logger.info(f"{Fore.GREEN}Disconnected HISTORICAL TrueData...{Style.RESET_ALL}")
-            print(f"{Style.NORMAL}{Fore.BLUE}Disconnected from Historical Data REST Connection !{Style.RESET_ALL}")
+            self.logger.warning(f"{Style.NORMAL}{Fore.BLUE}Disconnected from Historical Data REST Connection !{Style.RESET_ALL}")
 
     @staticmethod
     def truedata_duration_map(regular_format, end_date):
@@ -145,10 +149,8 @@ class TD:
                                                                    bar_size,
                                                                    options=options,
                                                                    bidask=bidask)
-        if ticker_id is None:
-            ticker_id = self.default_historical_data_id
-            self.default_historical_data_id = self.default_historical_data_id + 1
-        self.hist_data[ticker_id] = hist_data
+        if ticker_id is not None:
+            self.hist_data[ticker_id] = hist_data
         return hist_data
 
     def get_historical_data_from_duration(self, contract,
@@ -167,10 +169,8 @@ class TD:
         start_time = start_time.strftime('%y%m%d') + 'T00:00:00'    # This is the request format
 
         hist_data = self.historical_datasource.get_historic_data(contract, end_time, start_time, bar_size, options=options, bidask=bidask)
-        if ticker_id is None:
-            ticker_id = self.default_historical_data_id
-            self.default_historical_data_id = self.default_historical_data_id + 1
-        self.hist_data[ticker_id] = hist_data
+        if ticker_id is not None:
+            self.hist_data[ticker_id] = hist_data
         return hist_data
 
     def get_historical_data_from_start_time(self, contract,
@@ -188,10 +188,8 @@ class TD:
         end_time = end_time.strftime('%y%m%dT%H:%M:%S')    # This is the request format
         start_time = start_time.strftime('%y%m%dT%H:%M:%S')    # This is the request format
         hist_data = self.historical_datasource.get_historic_data(contract, end_time, start_time, bar_size, options=options, bidask=bidask)
-        if ticker_id is None:
-            ticker_id = self.default_historical_data_id
-            self.default_historical_data_id = self.default_historical_data_id + 1
-        self.hist_data[ticker_id] = hist_data
+        if ticker_id is not None:
+            self.hist_data[ticker_id] = hist_data
         return hist_data
 
     def get_bhavcopy(self, segment: str, date: datetime = None, return_completed: bool = True) -> List[Dict]:
@@ -244,5 +242,29 @@ class TD:
     def stop_live_data(self, contracts):  # Clearing objects is done after server confirmation
         self.live_websocket.send(f'{{"method": "removesymbol", "symbols": {json.dumps(contracts)}}}')
 
+    def trade_callback(self, func: Callable[[TickLiveData], Any]):
+        self.logger.info(f"Defining {func} as trade_callback...")
+        self.live_websocket.trade_callback = func
+
+    def clear_trade_callback(self):
+        self.logger.info(f"Clearing trade_callback...")
+        self.live_websocket.trade_callback = None
+
+    def bidask_callback(self, func: Callable[[TickLiveData], Any]):
+        self.logger.info(f"Defining bidask_callback...")
+        self.live_websocket.bidask_callback = func
+
+    def clear_bidask_callback(self):
+        self.logger.info(f"Clearing bidask_callback...")
+        self.live_websocket.bidask_callback = None
+
+    def bar_callback(self, func: Callable[[TickLiveData], Any]):
+        self.logger.info(f"Defining bar_callback...")
+        self.live_websocket.bar_callback = func
+
+    def clear_bar_callback(self):
+        self.logger.info(f"Clearing bar_callback...")
+        self.live_websocket.bar_callback = None
+
     # def get_touchline(self):
     #     self.live_websocket.send(f'{{"method": "touchline"}}')
